#ifdef _WIN32
#include <windows.h>
#else
#include "../../WDL/swell/swell.h"
#include "../../WDL/swell/swell-dlggen.h"
#endif

#include <stdio.h>
#include <math.h>
#include <sys/stat.h>

#include "resource.h"

#include "mp3dec.h"

// todo: lame gapless support?

//#include "main.h"
//#include "resource.h"
#include "../reaper_plugin.h"
#include "../../WDL/LineParse.h"
#include "../../WDL/wdlstring.h"
#include "../../WDL/ptrlist.h"
#include "../../WDL/mutex.h"

#include "../../WDL/fileread.h"
#include "../../WDL/filewrite.h"


#define WDL_WIN32_UTF8_IMPL static
#include "../../WDL/win32_utf8.c"


REAPER_PLUGIN_HINSTANCE g_hInst;

HWND g_main_hwnd;

REAPER_Resample_Interface *(*Resampler_Create)();
void (*format_timestr)(double tpos, char *buf, int buflen);
REAPER_PeakGet_Interface *(*PeakGet_Create)(const char *fn, int srate, int nch);
REAPER_PeakBuild_Interface *(*PeakBuild_Create)(PCM_source *src, const char *fn, int srate, int nch);
void (*resolve_fn)(const char *in, char *out, int outlen);
void (*relative_fn)(const char *in, char *out, int outlen);
void (*GetPeakFileName)(const char *fn, char *buf, int bufmax);
void (*update_disk_counters)(int read, int write);
const char *(*get_ini_file)();
void (*GetPreferredDiskReadMode)(int *mode, int *nb, int *bs);
void (*GetPreferredDiskReadModePeak)(int *mode, int *nb, int *bs);
void (*HiresPeaksFromSource)(PCM_source *src, PCM_source_peaktransfer_t *block);

#define SOURCE_TYPE "MP3"

#define POOLEDSRC_ALLOWPOOLMETASAVE 2

#include "mp3_index.h"





#include "../../WDL/ptrlist.h"
#include "../../WDL/rpool.h"


class PooledDecoderInstance
{
public:
  PooledDecoderInstance() 
  {
    m_resampler=0;
    m_resampler_initted=0;
    m_lastpos = -10000000; 
    m_decode_srcsplpos=0;
    m_file=0;
    m_dump_samples=0;
    m_stream_pos=0;
    m_need_initial_dump=true;
  }
  ~PooledDecoderInstance() 
  {
    delete m_resampler;
    delete m_file;
  }

  WDL_ResourcePool_ResInfo m_rpoolinfo;

  REAPER_Resample_Interface *m_resampler;
  bool m_resampler_initted;

  INT64 m_lastpos,m_decode_srcsplpos;
  mp3_decoder m_decoder;    
  unsigned int m_stream_pos;
  int m_dump_samples;
  bool m_need_initial_dump;
  WDL_FileRead *m_file;

};


class PooledDecoderMediaInfo
{
public:
  const char *m_err;
  PooledDecoderMediaInfo() 
  {
    m_err=0;
    memset(&m_syncframeinfo,0,sizeof(m_syncframeinfo));
    m_hiressrc=0;
    m_index=NULL;
    m_lengthsamples=0;
    m_stream_start=m_stream_len=0;
    m_srate=0;
    m_nch=0;
  }
  ~PooledDecoderMediaInfo() 
  {
    delete m_hiressrc;
    if (m_index)
    {
      mp3_index::release_index(m_index);
      m_index=0;
    }
  }
  int GetNumChannels() { return m_nch; } // return number of channels
  double GetSampleRate() { return m_srate; }

  INT64 GetLengthSamples(bool wantLatencyAdjust)
  {
    INT64 ls = m_lengthsamples;
    if (wantLatencyAdjust && m_index)
      ls -= m_index->m_start_eatsamples + m_index->m_end_eatsamples;
    return ls;
  }

  double GetLength(bool wantLatencyAdjust)  // length in seconds
  { 
    if (m_srate<1) return 0.0;
    return GetLengthSamples(wantLatencyAdjust) / (double)m_srate;
  } 
  double GetPreferredPosition() { return -1.0; }
  int GetBitsPerSample() { return 16; }

  bool IsAvailable() { return m_srate>0 && m_nch; }

  PooledDecoderInstance *Open(const char *filename, int peaksMode, bool forceread)
  {
    int mode=1,bs=8192, nb=2;
    if (peaksMode)
    {
      mode=0;
      nb=1;
      bs=8192;
      if (GetPreferredDiskReadModePeak) GetPreferredDiskReadModePeak(&mode,&nb,&bs);
    }
    else
    {
      if (GetPreferredDiskReadMode) GetPreferredDiskReadMode(&mode,&nb,&bs);
    }
    WDL_FileRead *file = new WDL_FileRead(filename,mode,min(bs,32768),min(nb,4));

    if (!file->IsOpen()) 
    {
      m_err="Error opening file";
      delete file;
      return NULL;
    }

    PooledDecoderInstance *pdi=new PooledDecoderInstance;
    int pos=0;
    if (!m_index || forceread || !m_srate || 
        !m_nch || !m_syncframeinfo.framesize || m_err) // sync
    {
      file->SetPosition(file->GetSize()-128);
      char id3[128];
      file->Read(id3,128);
      m_stream_len = (unsigned int)file->GetPosition();
      if (!memcmp(id3,"TAG",3)) m_stream_len -= 128;

      m_stream_start=0;

      if (!m_index)
        m_index = mp3_index::indexFromFilename(filename);

      if (m_index && m_index->GetFrameCount())
        m_stream_start = m_index->GetFramePos(0);

      file->SetPosition(m_stream_start);

      m_stream_len-=m_stream_start;

      while (!pdi->m_decoder.SyncState())
      {
        if (pdi->m_decoder.queue_bytes_in.Available()<4096)
        {
          char buf[4096];
          int l=m_stream_len-pos;
          if (l > sizeof(buf)) l=sizeof(buf);
          l=file->Read(buf,l);
          if (!l) break;

          pos+=l;
          if (l > 0) pdi->m_decoder.queue_bytes_in.Add(buf,l);
        }
        if (pdi->m_decoder.Run()) break;
      }
      pdi->m_decoder.queue_bytes_in.Compact();

      if (pdi->m_decoder.SyncState())
      {
        m_srate = pdi->m_decoder.GetSampleRate();
        m_nch = pdi->m_decoder.GetNumChannels();
        if (m_index) m_lengthsamples = (INT64) ((m_index->GetFrameCount()*(double)m_srate) / pdi->m_decoder.GetFrameRate() + 0.5);
        else m_lengthsamples=0;    
        m_syncframeinfo = pdi->m_decoder.m_lastframe;
      }
      else 
      {
        m_err="Error synchronizing to MPEG bitstream";
        delete pdi;
        delete file;
        return NULL;
      }
    }
    else
    {
      pos = m_stream_start;      
      pdi->m_decoder.m_lastframe = m_syncframeinfo;
    }
    m_err=0;
    pdi->m_file = file;
    pdi->m_stream_pos = pos;

    return pdi;
  }

  PCM_source *m_hiressrc; // can be shared across all instances since it's singlethreaded and without state

  int m_srate,m_nch;
  INT64 m_lengthsamples;
  unsigned int m_stream_start, m_stream_len;
  mp3_index *m_index;
  struct frame m_syncframeinfo;
};


#define POOLED_PCMSOURCE_CLASSNAME PCM_source_mp3

#define POOLED_PCMSOURCE_EXTRASTUFF_GETLENGTH_PARM (m_adjustLatency)
#define POOLED_PCMSOURCE_EXTRASTUFF_INIT m_adjustLatency=true;
#define POOLED_PCMSOURCE_EXTRASTUFF bool m_adjustLatency;
#define POOLED_PCMSOURCE_EXTRASTUFF_DUPLICATECODE ns->m_adjustLatency = m_adjustLatency; 

#include "../pooled_pcmsource_impl.h"

int POOLED_PCMSOURCE_CLASSNAME::PoolExtended(int call, void* parm1, void* parm2, void* parm3)
{
  return 0;
}

PooledDecoderInstance *POOLED_PCMSOURCE_CLASSNAME::CreateInstance()
{
  if (!m_filepool->extraInfo) m_filepool->extraInfo = new PooledDecoderMediaInfo;
  return m_filepool->extraInfo->Open(GetFileName(),m_peaksMode,!m_filepool->HasResources());
}



void PCM_source_mp3::PooledGetSamples(PCM_source_transfer_t *block, PooledDecoderInstance *poolreadinst, void *rsModeToUse)
{
  if (!poolreadinst->m_file) return;
  // process from file
  if (poolreadinst->m_rpoolinfo.m_ownerptr != this)
  {
    poolreadinst->m_lastpos=-1000; // force a reset of any SRC
  }

  int do_resample=0;
  ReaSample *sampleoutptr=block->samples;
  int len=block->length;
  double lat=0.0;
  double decsr=m_filepool->extraInfo->m_srate;
  if (fabs(decsr-block->samplerate)>=0.00001)
  {
    if (!poolreadinst->m_resampler_initted && !poolreadinst->m_resampler)
    {
      poolreadinst->m_resampler_initted=true;
      poolreadinst->m_resampler=Resampler_Create();
    }
    if (poolreadinst->m_resampler)
    {
      do_resample=1;
      poolreadinst->m_resampler->Extended(RESAMPLE_EXT_SETRSMODE,rsModeToUse,NULL,NULL);
      poolreadinst->m_resampler->SetRates(decsr,block->samplerate);
      lat=poolreadinst->m_resampler->GetCurrentLatency();
    }
  }

  if (block->absolute_time_s==-100000.0) poolreadinst->m_lastpos=-100000;

  if (fabs(poolreadinst->m_lastpos - (block->time_s+lat)*decsr) >= 8.0)
  {
    mp3_index *mindex = m_filepool->extraInfo->m_index;
    INT64 splpos = poolreadinst->m_decode_srcsplpos = (INT64) (block->time_s * decsr + 0.5);

    if (mindex && mindex->GetFrameCount() > 0)
    {
      int framesize=(decsr < 32000 ? 576 : 1152);

      int frame_pos = (int)(splpos / framesize) - 10; // seek ahead
      if (frame_pos < 0) frame_pos=0;
      else if (frame_pos >= mindex->GetFrameCount()) frame_pos=mindex->GetFrameCount()-1;

      poolreadinst->m_stream_pos=mindex->GetFramePos(frame_pos) - m_filepool->extraInfo->m_stream_start;

      poolreadinst->m_dump_samples = (int) (splpos - (((INT64)frame_pos)*framesize)); 

      if (m_adjustLatency) poolreadinst->m_dump_samples += mindex->m_start_eatsamples;
      else
      {
        // ugh, old code was so inaccurate, this should make it match
        poolreadinst->m_dump_samples -= framesize; // insert an extra frame of latency  (old seek code had an excess +1 in the frame count)
        if (poolreadinst->m_dump_samples<0)poolreadinst->m_dump_samples=0;
      }

      poolreadinst->m_need_initial_dump=false;
    }
    else
    {
      poolreadinst->m_dump_samples=0;
      poolreadinst->m_stream_pos=(int) ((m_filepool->extraInfo->m_stream_len) * block->time_s * decsr / m_filepool->extraInfo->GetLengthSamples(m_adjustLatency));
    }
    poolreadinst->m_file->SetPosition(m_filepool->extraInfo->m_stream_start + poolreadinst->m_stream_pos);

    poolreadinst->m_decoder.Reset(false);
    poolreadinst->m_lastpos=(INT64) (block->time_s*decsr);
    if (poolreadinst->m_resampler) poolreadinst->m_resampler->Reset();
    lat=0.0;
  }

  if (poolreadinst->m_need_initial_dump)
  {
    mp3_index *mindex = m_filepool->extraInfo->m_index;

    if (m_adjustLatency && mindex) poolreadinst->m_dump_samples += mindex->m_start_eatsamples;

    poolreadinst->m_need_initial_dump=false;
  }

  if (do_resample)
  {
    len = poolreadinst->m_resampler->ResamplePrepare(block->length,block->nch,&sampleoutptr);
  }

  if (poolreadinst->m_dump_samples<0)
  {
    int l=(-poolreadinst->m_dump_samples)*sizeof(double)*poolreadinst->m_decoder.GetNumChannels();
    poolreadinst->m_dump_samples=0;
    void *b=poolreadinst->m_decoder.queue_samples_out.Add(NULL,l);
    memset(b,0,l);
  }

  int tr=0;
  int hasHadRdError=0;
  while (poolreadinst->m_decoder.queue_samples_out.Available() < len*(int)sizeof(double)*poolreadinst->m_decoder.GetNumChannels())
  {
    if (poolreadinst->m_decoder.queue_bytes_in.Available() < 4096)
    {
      int l=m_filepool->extraInfo->m_stream_len-poolreadinst->m_stream_pos;
      char buf[4096];
      if (l > sizeof(buf)) l=sizeof(buf);
      l=poolreadinst->m_file->Read(buf,l);
      if (l<1) hasHadRdError=true;
      else
      {
        poolreadinst->m_decoder.queue_bytes_in.Add(buf,l);
        tr+=l;
        poolreadinst->m_stream_pos+=l;
      }
    }


    int os=poolreadinst->m_decoder.queue_samples_out.Available();
    if (poolreadinst->m_decoder.Run()) break;
    int l=poolreadinst->m_decoder.queue_samples_out.Available();

    if (l <= os && hasHadRdError) break;

    if (poolreadinst->m_dump_samples>0 && l > 0)
    {
      l /= sizeof(double) * poolreadinst->m_decoder.GetNumChannels();
      if (l > poolreadinst->m_dump_samples) l=poolreadinst->m_dump_samples;
      poolreadinst->m_decoder.queue_samples_out.Advance(l*sizeof(double)*poolreadinst->m_decoder.GetNumChannels());
      poolreadinst->m_dump_samples -= l;
      if (poolreadinst->m_dump_samples<0) poolreadinst->m_dump_samples=0;
    }
  }
  poolreadinst->m_decoder.queue_bytes_in.Compact();

  if (tr&&update_disk_counters) update_disk_counters(tr,0);

  int samples_read=0;
  if (poolreadinst->m_decoder.GetNumChannels()) 
  {
    samples_read=poolreadinst->m_decoder.queue_samples_out.Available()/sizeof(double)/poolreadinst->m_decoder.GetNumChannels();

    INT64 maxs  = (m_filepool->extraInfo->GetLengthSamples(m_adjustLatency) - poolreadinst->m_decode_srcsplpos);

    if (maxs<0)maxs=0;
    if (!m_adjustLatency && samples_read > maxs) samples_read = (int)maxs;
    if (samples_read > len) samples_read=len;
    poolreadinst->m_decode_srcsplpos+=samples_read;

  }

  if (samples_read > 0)
  {
    poolreadinst->m_lastpos += samples_read;
    // copy the samples to sampleoutptr, converting to stereo if necessary
    if (poolreadinst->m_decoder.GetNumChannels() == block->nch)
    {
      memcpy(sampleoutptr,poolreadinst->m_decoder.queue_samples_out.Get(),samples_read*sizeof(double)*block->nch);
      poolreadinst->m_decoder.queue_samples_out.Advance(samples_read*sizeof(double)*block->nch);
    }
    else if (poolreadinst->m_decoder.GetNumChannels() == 1)
    {
      double *inptr=(double *)poolreadinst->m_decoder.queue_samples_out.Get();
      ReaSample *outptr=sampleoutptr;
      int i;
      for (i = 0; i < samples_read; i ++)
      {
        double s=*inptr++;
        int ch;
        for (ch = 0; ch < block->nch; ch ++)
          *outptr++ = s;
      }
      poolreadinst->m_decoder.queue_samples_out.Advance(samples_read*sizeof(double));
    }
    else if (poolreadinst->m_decoder.GetNumChannels() == 2)
    {
      double *inptr=(double *)poolreadinst->m_decoder.queue_samples_out.Get();
      ReaSample *outptr=sampleoutptr;
      int i;
      for (i = 0; i < samples_read; i ++)
      {
        double s=inptr[0] * 0.5 + inptr[1] * 0.5;
        inptr+=2;
        int ch;
        for (ch = 0; ch < block->nch; ch ++)
          *outptr++ = s;
      }
      poolreadinst->m_decoder.queue_samples_out.Advance(samples_read*sizeof(double)*2);
    }
    poolreadinst->m_decoder.queue_samples_out.Compact();
  }

  if (!do_resample)
  {
    block->samples_out = samples_read;
  }
  else // resample it now
  {
    block->samples_out=poolreadinst->m_resampler->ResampleOut(block->samples,samples_read, block->length,block->nch);
  }

  // output to block

}

WDL_DLGRET PCM_source_mp3::_dlgProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == WM_INITDIALOG) SetWindowLongPtr(hwndDlg,GWLP_USERDATA,lParam);
  PCM_source_mp3 *_this = (PCM_source_mp3 *)GetWindowLongPtr(hwndDlg,GWLP_USERDATA);
  return _this->propsDlgProc(hwndDlg,uMsg,wParam,lParam);
}


int PCM_source_mp3::PropertiesWindow(HWND hwndParent)
{
  return DialogBoxParam(g_hInst,MAKEINTRESOURCE(IDD_WAVEITEMINFO),hwndParent,_dlgProc,(LPARAM)this);
}


WDL_DLGRET PCM_source_mp3::propsDlgProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  switch (uMsg)
  {
    case WM_INITDIALOG:
      SetDlgItemText(hwndDlg,IDC_NAME,GetFileName());

      CheckDlgButton(hwndDlg,IDC_CHECK1,m_adjustLatency?BST_CHECKED:BST_UNCHECKED);
      if (!IsAvailable())
      {
        if (!m_filepool||!m_filepool->extraInfo)
        {
          SetDlgItemText(hwndDlg,IDC_INFO,m_isOffline?"File offline":"File not opened");
        }
        else if (m_filepool->extraInfo->m_err)
          SetDlgItemText(hwndDlg,IDC_INFO,m_filepool->extraInfo->m_err);
        else 
          SetDlgItemText(hwndDlg,IDC_INFO,"Unknown error/status");
      }
      else
      {

        char buf[512];
        WDL_String s;
        PooledDecoderInstance *inst=GetFile();
        if (inst)
        {
          sprintf(buf,"MPEG layer %d, @ %d Hz %d channels\r\n",inst->m_decoder.GetLayer(), inst->m_decoder.GetSampleRate(),inst->m_decoder.GetNumChannels());
          s.Append(buf);
          ReleaseFile(inst,0);
        }

        if (m_filepool && m_filepool->extraInfo && m_filepool->extraInfo->m_index && m_filepool->extraInfo->m_index->GetFrameCount())
        {
          s.Append("Length: ");
          format_timestr(m_filepool->extraInfo->GetLengthSamples(m_adjustLatency) / (double)m_filepool->extraInfo->m_srate,buf,sizeof(buf));
          s.Append(buf);
          sprintf(buf,"\r\nBitrate (average): %.0f kbps\r\n",(double)m_filepool->extraInfo->m_index->GetFramePos(m_filepool->extraInfo->m_index->GetFrameCount()-1) / (m_filepool->extraInfo->GetLengthSamples(m_adjustLatency)/(double)m_filepool->extraInfo->m_srate) * 8.0/1000.0);
          s.Append(buf);
          char buf[512];
          sprintf(buf,"%d frames in file [indexed]\r\n",m_filepool->extraInfo->m_index->GetFrameCount());
          s.Append(buf);
        }
        else
        {
          s.Append("No index found [error]!\r\n");
        }
        SetDlgItemText(hwndDlg,IDC_INFO,s.Get());
      }

    return 0;
    case WM_COMMAND:
      switch (LOWORD(wParam))
      {
        case IDOK:
        case IDCANCEL:
          m_adjustLatency = !!IsDlgButtonChecked(hwndDlg,IDC_CHECK1);
          EndDialog(hwndDlg,0);
        return 0;
      }
    return 0;
  }
  return 0;
}


void PCM_source_mp3::SaveState(ProjectStateContext *ctx)
{
  char buf[2048];
  if (relative_fn)
  {
    relative_fn(GetFileName(),buf,sizeof(buf));
    ctx->AddLine("FILE \"%s\" %d",buf,m_adjustLatency);
  }
  else ctx->AddLine("FILE \"%s\" %d",GetFileName(),m_adjustLatency);
}

int PCM_source_mp3::LoadState(char *firstline, ProjectStateContext *ctx) // -1 on error
{
  int child_count=1;
  bool comment_state=false;
  m_adjustLatency=false;

  for (;;)
  {
    char linebuf[4096];
    if (ctx->GetLine(linebuf,sizeof(linebuf))) break;
    LineParser lp(comment_state);
    if (lp.parse(linebuf)||lp.getnumtokens()<=0) continue;
    if (child_count == 1)
    {
      if (lp.gettoken_str(0)[0] == '>') return 0;
      
      if (lp.gettoken_str(0)[0] == '<') child_count++;
      else if (!stricmp(lp.gettoken_str(0),"FILE"))
      {
        if (lp.getnumtokens()>1)
        {
          if (!IsAvailable() || stricmp(lp.gettoken_str(1),GetFileName()))
          {
            char buf[4096];
            resolve_fn(lp.gettoken_str(1),buf,sizeof(buf));
            Open(buf);
          }
          if (lp.getnumtokens()>2)
            m_adjustLatency = !!lp.gettoken_int(2);
        }
      }
    }
    else if (lp.gettoken_str(0)[0] == '<') child_count++;
    else if (lp.gettoken_str(0)[0] == '>') child_count--;
  }
  return -1;
}












PCM_source *CreateFromType(const char *type, int priority)
{
  if (priority > 4) // lowest priority wav reader
  {
    if (!strcmp(type,SOURCE_TYPE))
      return new PCM_source_mp3;
  }

  return NULL;
}

PCM_source *CreateFromFile(const char *filename, int priority)
{
  int lfn=strlen(filename);
  if (priority > 4 && lfn>4 && (!stricmp(filename+lfn-4,".mp3") || !stricmp(filename+lfn-4,".mp2")))
  {
    PCM_source_mp3 *w=new PCM_source_mp3;
    w->Open(filename);
    if (w->IsAvailable()) return w;
    delete w;
  }
  return NULL;
}

  // this is used for UI only, not so muc
const char *EnumFileExtensions(int i, char **descptr) // call increasing i until returns a string, if descptr's output is NULL, use last description
{
  if (i == 0)
  {
    if (descptr) *descptr = "MPEG Audio Files";
    return "MP2;MP3";
  }
  if (descptr) *descptr=NULL;

  if (i == 4096) return "REAPINDEX"; // let it know this shit should be deleted, if it can be
  return NULL;
}




pcmsrc_register_t myRegStruct={CreateFromType,CreateFromFile,EnumFileExtensions};

extern pcmsink_register_ext_t mySinkRegStruct;



#include "impegdecode.h"

class mpeg_decoder_inst : public IMPEGDecoder
{
public:
  mpeg_decoder_inst() { }
  virtual ~mpeg_decoder_inst() { }
  
  bool WantInput() 
  {
    return m_dec.queue_bytes_in.Available() < 4096;
  }
  void AddInput(const void *buf, int buflen)
  {
    m_dec.queue_bytes_in.Add(buf,buflen);
  }

  double *GetOutput(int *avail) // avail returns number of doubles avail
  {
    *avail = m_dec.queue_samples_out.Available()/sizeof(double);
    return (double *)m_dec.queue_samples_out.Get();
  }
  void OutputAdvance(int num)  // num = number of doubles
  {
    m_dec.queue_samples_out.Advance(num*sizeof(double));
    m_dec.queue_samples_out.Compact();
  }

  void Reset(bool full)
  {
    m_dec.Reset(full);
  }
  int SyncState() { return m_dec.SyncState(); } // > 0 if synched
  int Run() { int ret=m_dec.Run();  m_dec.queue_bytes_in.Compact(); return ret; } // returns -1 on non-recoverable error
  int GetSampleRate() { return m_dec.GetSampleRate(); }
  int GetNumChannels() { return m_dec.GetNumChannels(); }

  mp3_decoder m_dec;
};


IMPEGDecoder *CreateMPEGdecoder()
{
  return (IMPEGDecoder*) (new mpeg_decoder_inst);
}



extern "C"
{

REAPER_PLUGIN_DLL_EXPORT int REAPER_PLUGIN_ENTRYPOINT(REAPER_PLUGIN_HINSTANCE hInstance, reaper_plugin_info_t *rec)
{
  g_hInst=hInstance;
  if (rec)
  {
    if (rec->caller_version != REAPER_PLUGIN_VERSION || !rec->GetFunc)
      return 0;

    g_main_hwnd=rec->hwnd_main;
    *((void **)&Resampler_Create) = rec->GetFunc("Resampler_Create");
    *((void **)&format_timestr) = rec->GetFunc("format_timestr");
    *((void **)&PeakGet_Create) = rec->GetFunc("PeakGet_Create");
    *((void **)&PeakBuild_Create) = rec->GetFunc("PeakBuild_Create");
    *((void **)&resolve_fn) = rec->GetFunc("resolve_fn");   
    *((void **)&relative_fn) = rec->GetFunc("relative_fn");   
    *((void **)&GetPeakFileName) = rec->GetFunc("GetPeakFileName");    
    *((void **)&update_disk_counters) = rec->GetFunc("update_disk_counters");
    *((void **)&get_ini_file) = rec->GetFunc("get_ini_file");
    *((void **)&GetPreferredDiskReadMode) = rec->GetFunc("GetPreferredDiskReadMode");
    *((void **)&GetPreferredDiskReadModePeak) = rec->GetFunc("GetPreferredDiskReadModePeak");
    *((void **)&HiresPeaksFromSource) = rec->GetFunc("HiresPeaksFromSource");

    if (!PeakGet_Create || !PeakBuild_Create || !Resampler_Create || !format_timestr || !resolve_fn ||       
        !rec->Register)
          return 0;

    rec->Register("pcmsrc",&myRegStruct);
    rec->Register("API_CreateMPEGdecoder",(void*)CreateMPEGdecoder);

    if (!rec->Register("pcmsink_ext",&mySinkRegStruct))
      rec->Register("pcmsink",&mySinkRegStruct);

    // our plugin registered, return success

    return 1;
  }
  else
  {
    return 0;
  }
}

};


#ifndef _WIN32 // MAC resources
#include "../../WDL/swell/swell-dlggen.h"
#include "res.rc_mac_dlg"
#undef BEGIN
#undef END
#include "../../WDL/swell/swell-menugen.h"
#include "res.rc_mac_menu"
#endif
