#include "../WDL/rpool.h"
#include "../WDL/poollist.h"

#define DECODER_POOL_OWNAFTERREAD 3000

static WDL_PoolList<WDL_ResourcePool<PooledDecoderInstance,PooledDecoderMediaInfo> > g_file_pool[3]; // [m_peaksMode]

class FileNameListEnt
{
public:
  FileNameListEnt(char *identstr) { WDL_POOLLIST_identstr=identstr; WDL_POOLLIST_refcnt=0; }
  ~FileNameListEnt() {}  // do NOT free or delete WDL_POOLLIST_identstr

  void Clear() {}  // will be called if ReleasePool(x,false) is called and refcnt gets to 0
  int WDL_POOLLIST_refcnt;
  char *WDL_POOLLIST_identstr;
};

static WDL_PoolList<FileNameListEnt> s_filename_list;


class POOLED_PCMSOURCE_CLASSNAME : public PCM_source
{
  public:
    POOLED_PCMSOURCE_CLASSNAME()
    {
      m_filename_ent=0;
      m_isavail=false;
      m_isOffline=0;
      m_filepool=NULL;
      m_peakbuilder=0;
      m_peakgetter=0;
      m_peaksMode=0;
      m_resampler_rsmode = (void *)(INT_PTR)-1;
#ifdef POOLEDSRC_USE_SLICES
      m_slice=-1;
#endif
#ifdef POOLED_PCMSOURCE_EXTRASTUFF_INIT
      POOLED_PCMSOURCE_EXTRASTUFF_INIT
#endif
      Close();      
    }
#ifdef POOLEDSRC_USE_SLICES
    int m_slice;
#endif

    virtual ~POOLED_PCMSOURCE_CLASSNAME() 
    {
      Close();
      if (m_filename_ent)
      {
        s_filename_list.Release(m_filename_ent);
        m_filename_ent=0;
      }
    }

    PCM_source *Duplicate()
    {
      POOLED_PCMSOURCE_CLASSNAME *ns=new POOLED_PCMSOURCE_CLASSNAME;
#ifdef POOLEDSRC_USE_SLICES
      ns->m_slice = m_slice;
#endif
      if (m_filename_ent) ns->Open(m_filename_ent->WDL_POOLLIST_identstr);
#ifdef POOLED_PCMSOURCE_EXTRASTUFF_DUPLICATECODE
      POOLED_PCMSOURCE_EXTRASTUFF_DUPLICATECODE
#endif
      return ns;
    }
    const char *GetFileName() 
    { 
      return m_filename_ent && m_filename_ent->WDL_POOLLIST_identstr ? m_filename_ent->WDL_POOLLIST_identstr : ""; 
    }
    bool SetFileName(const char *newfn) 
    { 
      if (m_filename_ent)
      {
        s_filename_list.Release(m_filename_ent);
        m_filename_ent=0;
      }
      m_filename_ent=s_filename_list.Get(newfn);
      return true; 
    } 


    const char *GetType() { return SOURCE_TYPE; }
    int GetNumChannels() { return m_filepool && m_filepool->extraInfo ? m_filepool->extraInfo->GetNumChannels() : 1; } // return number of channels

    double GetPreferredPosition() 
    { 
      return m_filepool && m_filepool->extraInfo ? m_filepool->extraInfo->GetPreferredPosition() : -1.0; 
    } 

    double GetSampleRate() { return m_filepool && m_filepool->extraInfo ? m_filepool->extraInfo->GetSampleRate() : 0.0; }

    double GetLength()  // length in seconds
    { 
      return m_filepool && m_filepool->extraInfo ? m_filepool->extraInfo->GetLength(
#ifdef POOLEDSRC_USE_SLICES
        m_slice
#endif
#ifdef POOLED_PCMSOURCE_EXTRASTUFF_GETLENGTH_PARM
        POOLED_PCMSOURCE_EXTRASTUFF_GETLENGTH_PARM
#endif
        ) : 0;
    } 
#ifdef POOLEDSRC_USE_SLICES
    double GetLengthBeats();
#endif

    int GetBitsPerSample() { return m_filepool && m_filepool->extraInfo ? m_filepool->extraInfo->GetBitsPerSample() : 0; }

    bool IsAvailable() 
    { 
      return m_isavail && m_filepool && m_filepool->extraInfo && m_filepool->extraInfo->IsAvailable(); 
    }

    void SetAvailable(bool avail) 
    { 
      if (!avail)
      {
        m_isOffline=1;
        if (IsAvailable()) 
        {
          Close(false);
        }
      }
      else
      {
        m_isOffline=0;
        if (!IsAvailable())
          Open(NULL);
      }
   
    } // optional, if called with avail=false, close files/etc, and so on


    void OpenPeaks()
    {
#ifndef POOLEDSRC_GETPEAKINFO_NOIMPL
      if (!m_filepool||!m_filepool->extraInfo)
      {
        delete m_peakgetter;
        m_peakgetter=0;
        return;
      }


      if (!m_peakgetter && m_filename_ent) 
      {
        int sr=0,nch=0;
        if (m_filepool&&m_filepool->extraInfo)
        {
          sr = (int)m_filepool->extraInfo->GetSampleRate();
          nch = m_filepool->extraInfo->GetNumChannels();
        }
        if (sr>0 && nch>0)
          m_peakgetter=PeakGet_Create(m_filename_ent->WDL_POOLLIST_identstr,sr,nch);
      }
#endif
    }


    void Open(const char *filename, int peaksMode=0);
    void Close(bool full=true)
    {
#ifndef POOLEDSRC_ALLOWPOOLMETASAVE
      full=true;
#else
#if POOLEDSRC_ALLOWPOOLMETASAVE == 2
      if (m_filepool && m_filepool->extraInfo)
      {
        delete m_filepool->extraInfo->m_hiressrc;
        m_filepool->extraInfo->m_hiressrc=0;
      }
#endif
#endif
      g_file_pool[m_peaksMode].Release(m_filepool,full);
      m_filepool=0;

      delete m_peakbuilder;
      m_peakbuilder=0;

      delete m_peakgetter;
      m_peakgetter=0;

      m_isavail=false;
    }


    int PropertiesWindow(HWND hwndParent);

    void PooledGetSamples(PCM_source_transfer_t *block, PooledDecoderInstance *poolreadinst, void *rsModeToUse);
    void GetSamples(PCM_source_transfer_t *block)
    {
      block->samples_out=0;

      PooledDecoderInstance *poolreadinst=0;
      if (m_filepool && m_filepool->extraInfo && m_filepool->extraInfo->IsAvailable() && (poolreadinst=GetFile())) 
      {
        PooledGetSamples(block,poolreadinst,m_resampler_rsmode);

        // output to block
        ReleaseFile(poolreadinst,m_peaksMode ? 0 : DECODER_POOL_OWNAFTERREAD);
      }
    }


    void SaveState(ProjectStateContext *ctx);
    int LoadState(char *firstline, ProjectStateContext *ctx); // -1 on error


    void GetPeakInfo(PCM_source_peaktransfer_t *block)
#ifdef POOLEDSRC_GETPEAKINFO_NOIMPL
      ;
#else
    {
      block->peaks_out=0;
      if (!m_filepool || !m_filepool->extraInfo) return;

      if (m_peakbuilder)
      {
        if (block->peakrate >= REAPER_PEAKRES_MAX_NOPKS)  
        {
          if (!m_filepool->extraInfo->m_hiressrc && m_filename_ent)
          {
            POOLED_PCMSOURCE_CLASSNAME *t=new POOLED_PCMSOURCE_CLASSNAME;
            t->Open(m_filename_ent->WDL_POOLLIST_identstr,2);
            m_filepool->extraInfo->m_hiressrc=t;
          }
          if (m_filepool->extraInfo->m_hiressrc)
          {
            if (HiresPeaksFromSource) HiresPeaksFromSource(m_filepool->extraInfo->m_hiressrc,block);
            if (block->peaks_out) return;
          }
        }
        m_peakbuilder->GetPeakInfo(block);
        return;
      }

      if (!m_filepool->extraInfo->GetNumChannels() || !block->numpeak_points) return;

      double peakres=m_peakgetter ? m_peakgetter->GetMaxPeakRes() : REAPER_PEAKRES_MAX_NOPKS;

      if (block->peakrate >= peakres*REAPER_PEAKRES_MUL_MAX)
      {
        if (!m_filepool->extraInfo->m_hiressrc && m_filename_ent)
        {
          POOLED_PCMSOURCE_CLASSNAME *t=new POOLED_PCMSOURCE_CLASSNAME;
          t->Open(m_filename_ent->WDL_POOLLIST_identstr,2);
          m_filepool->extraInfo->m_hiressrc=t;
        }
        if (m_filepool->extraInfo->m_hiressrc)
        {
          if (HiresPeaksFromSource) HiresPeaksFromSource(m_filepool->extraInfo->m_hiressrc,block);
          if (block->peaks_out) return;
        }
      }

      if (m_peakgetter) m_peakgetter->GetPeakInfo(block);
    }
#endif
    void Peaks_Clear(bool deleteFile) 
    { 
#ifndef POOLEDSRC_GETPEAKINFO_NOIMPL
      delete m_peakgetter;
      m_peakgetter=0;

      if (deleteFile && GetPeakFileName && m_filename_ent)
      {
        char peakfn[2048];
        GetPeakFileName(m_filename_ent->WDL_POOLLIST_identstr,peakfn,sizeof(peakfn));
        DeleteFile(peakfn);
      }
#endif
    }
    int PeaksBuild_Begin() 
    { 
#ifndef POOLEDSRC_GETPEAKINFO_NOIMPL
      OpenPeaks();

      if (m_peakgetter||!IsAvailable()||!m_filename_ent) return 0;

      delete m_peakbuilder;

      POOLED_PCMSOURCE_CLASSNAME *ni = new POOLED_PCMSOURCE_CLASSNAME;
      ni->Open(m_filename_ent->WDL_POOLLIST_identstr,1);
      int sr=0,nch=0;
      if (m_filepool&&m_filepool->extraInfo)
      {
        sr = (int)m_filepool->extraInfo->GetSampleRate();
        nch = m_filepool->extraInfo->GetNumChannels();
      }
      if (sr>0 && nch>0)
      {
        m_peakbuilder = PeakBuild_Create(ni,m_filename_ent->WDL_POOLLIST_identstr,sr,nch);
        return 1; 
      }
      return 0;

#else
      return 0;
#endif
    } // returns nonzero if building is opened, otherwise it may mean building isn't necessary
    int PeaksBuild_Run() 
    { 
#ifndef POOLEDSRC_GETPEAKINFO_NOIMPL
      if (m_peakbuilder)
      {
        return m_peakbuilder->Run();
      }
#endif
      return 0; 
    } // returns nonzero if building should continue

    void PeaksBuild_Finish() 
    { 
#ifndef POOLEDSRC_GETPEAKINFO_NOIMPL
      delete m_peakbuilder;
      m_peakbuilder=0;
      OpenPeaks();
#endif
    } // called when done


    int Extended(int call, void *parm1, void *parm2, void *parm3) 
    { 
      if (call == PCM_SOURCE_EXT_ENDPLAYNOTIFY)
      {
        if (m_filepool) m_filepool->ReleaseResources(this);
        
        return 1;
      }
      if (call == PCM_SOURCE_EXT_SETRESAMPLEMODE) 
      {
        m_resampler_rsmode=parm1;
        return 1;
      }
      if (m_filepool && m_filepool->extraInfo)
        return PoolExtended(call,parm1,parm2,parm3);
      return 0;
    }

    int PoolExtended(int call, void *parm1, void *parm2, void *parm3);


    FileNameListEnt *m_filename_ent;

    WDL_ResourcePool<PooledDecoderInstance, PooledDecoderMediaInfo> *m_filepool;

    char m_isOffline;
    int m_peaksMode;
  protected:

    REAPER_PeakGet_Interface *m_peakgetter;
    REAPER_PeakBuild_Interface *m_peakbuilder;

    void *m_resampler_rsmode;


  WDL_DLGRET propsDlgProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
  static WDL_DLGRET _dlgProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);


  bool m_isavail;

  PooledDecoderInstance *GetFile();
  void ReleaseFile(PooledDecoderInstance *f, int len);

  PooledDecoderInstance *CreateInstance();

#ifdef POOLED_PCMSOURCE_EXTRASTUFF
  POOLED_PCMSOURCE_EXTRASTUFF
#endif
};

void POOLED_PCMSOURCE_CLASSNAME::Open(const char *filename, int peaksMode)
{
  Close(!!filename);

  m_peaksMode=peaksMode;
  if (filename && strcmp(filename,GetFileName())) 
  {
    if (m_filename_ent)
    {
      s_filename_list.Release(m_filename_ent);
      m_filename_ent=0;
    }
    m_filename_ent=s_filename_list.Get(filename);
  }

  m_filepool=m_filename_ent ? g_file_pool[m_peaksMode].Get(m_filename_ent->WDL_POOLLIST_identstr) : 0;

  if (m_filepool)
  {
    if (!m_filepool->HasResources())
    {
      PooledDecoderInstance *poolreadinst=GetFile();
      if (poolreadinst) 
      {
        m_isavail=true;
        ReleaseFile(poolreadinst,0);
        if (!peaksMode) OpenPeaks();
      }
    }
    else 
    {
      m_isavail=true;
      if (!peaksMode) OpenPeaks();
    }
  }

}

void POOLED_PCMSOURCE_CLASSNAME::ReleaseFile(PooledDecoderInstance *f, int len)
{
  if (!m_filepool) delete f;
  else
  {
#ifdef _WIN32
    m_filepool->AddResource(f,this,len > 0 ? (GetTickCount()+len) : 0);
#else
    m_filepool->AddResource(f,this,len > 0 ? (time(NULL)+(len+999)/1000) : 0);
#endif
  }
}

PooledDecoderInstance *POOLED_PCMSOURCE_CLASSNAME::GetFile()
{
  if (!m_filepool) return 0;
#ifdef _WIN32
  PooledDecoderInstance *p=m_filepool->GetResource(this,GetTickCount());
#else
  PooledDecoderInstance *p=m_filepool->GetResource(this,time(NULL));
#endif
  if (p) return p;

  return CreateInstance();

}


/*

  
#include "../../WDL/ptrlist.h"
#include "../../WDL/rpool.h"


class PooledDecoderInstance
{
public:
  PooledDecoderInstance() 
  {
    m_resampler=0;
    m_resampler_initted=0;
    m_lastpos = -10000000; 
  }
  ~PooledDecoderInstance() 
  { 
    delete m_resampler;
  }

  WDL_ResourcePool_ResInfo m_rpoolinfo;

  REAPER_Resample_Interface *m_resampler;
  bool m_resampler_initted;
  INT64 m_lastpos;
};

class PooledDecoderMediaInfo // todo: metadata
{
public:
  PooledDecoderMediaInfo() 
  {
    m_hiressrc=0;
  }
  ~PooledDecoderMediaInfo() 
  {

    delete m_hiressrc;
  }

  PooledDecoderInstance *Open(const char *filename, int peaksMode, bool forceread)
  {
    PooledDecoderInstance *ndec = new PooledDecoderInstance;
    return ndec;
  }


  int GetNumChannels() { return m_nch; } // return number of channels
  double GetSampleRate() { return m_srate; }
  double GetLength()  // length in seconds
  { 
    if (m_srate<1.0) return 0; 
    return m_lengthsamples / m_srate; 
  } 
  int GetBitsPerSample() { return m_bps; }
  bool IsAvailable() 
  { 
    return m_srate>=1.0 && m_bps && m_nch; 
  }

  double GetPreferredPosition()
  {
    return -1.0;
  }


  PCM_source *m_hiressrc; // can be shared across all instances since it's singlethreaded and without state

};


#define POOLED_PCMSOURCE_CLASSNAME PCM_source_whatever

#include "../pooled_pcmsource_impl.h"

PooledDecoderInstance *POOLED_PCMSOURCE_CLASSNAME::CreateInstance()
{
  if (!m_filepool->extraInfo) m_filepool->extraInfo = new PooledDecoderMediaInfo;
  return m_filepool->extraInfo->Open(GetFileName(),m_peaksMode,!m_filepool->HasResources());
}

void POOLED_PCMSOURCE_CLASSNAME::PooledGetSamples(PCM_source_transfer_t *block, PooledDecoderInstance *poolreadinst)
{
}
int POOLED_PCMSOURCE_CLASSNAME::PoolExtended(int call, void* parm1, void* parm2, void* parm3)
{
  return 0;
}

  */